// -------------------------------------------------------------------------------- 
//!  @file  bg.c                                                                    
//!  背景タスク処理のソースファイル                                                 
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "enemy/enemy.h"
#include "bg.h"
#include "stage01.h"


// -------------------------------------------------------------------------------- 
//!  画面描画                                                                       
//!                                                                                 
//!  @param [in] p_data  データ                                                     
//!  @return             描画後のデータ                                             
// -------------------------------------------------------------------------------- 
static uint16_t const *drawScreen( uint16_t const *p_data )
{
	uint8_t		*p_attr ;
	uint16_t	*p_dest ;
	int			x, y ;


	for(x=0; x<(SCREEN_WIDTH+16)/8; x+=2) {
		p_attr = &g.bg.attr[0][x] ;
		p_dest = &g.vram.tile[0][x] ;

		p_data++ ;

		for(y=0; y<SCREEN_HEIGHT/8; y++) {

			// 属性
			p_attr[0] = (uint8_t )(p_data[0] >> 8) ;
			p_attr[1] = (uint8_t )(p_data[0] >> 0) ;
			p_attr   += 64 ;

			// タイル
			p_dest[0] = p_data[1] ;
			p_dest[1] = p_data[2] ;
			p_dest   += 64 ;

			p_data += 3 ;
		}
	}

	return &p_data[0] ;
}


// -------------------------------------------------------------------------------- 
//!  敵タスクを生成                                                                 
//!                                                                                 
//!  @param [in,out] p_tcb  背景タスク TCB                                          
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void RAMFUNC createEnemy( BGTCB *p_tcb )
{
	uint16_t	const *p_enemy ;
	uint32_t	exist ;
	int			y ;


	if( !p_tcb->enemyCount ) {
		return ;
	}

	if( --p_tcb->enemyCount ) {
		return ;
	}

	p_enemy = &p_tcb->p_enemy[0] ;
	exist   = ((uint32_t )p_enemy[1] << 16) | p_enemy[0] ;
	p_enemy += 2 ;

	y = 0 ;

	while( exist ) {
		if( exist & 1 ) {
			enemy_create(y*FP_ONE, *p_enemy++) ;
		}
		y      += 8 ;
		exist >>= 1 ;
	}

	p_tcb->enemyCount = *p_enemy++ ;
	p_tcb->p_enemy    = &p_enemy[0] ;
}


// -------------------------------------------------------------------------------- 
//!  背景 VBlank 処理                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  背景タスク TCB                                          
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void RAMFUNC bg_vblank( BGTCB *p_tcb )
{
	uint8_t		*p_attr ;
	uint16_t	*p_dest ;
	uint16_t	const *p_data ;
	int			x, y ;

	// １列描画
	x      = (getInt( p_tcb->x ) / 16) * 2 ;
	p_data = &p_tcb->p_data[1] ;
	p_attr = &g.bg.attr[0][(x+(SCREEN_WIDTH+16)/8)&63] ;
	p_dest = &g.vram.tile[0][(x+(SCREEN_WIDTH+16)/8)&63] ;

	for(y=0; y<SCREEN_HEIGHT/8; y++) {

		// 属性
		p_attr[0] = (uint8_t )(p_data[0] >> 8) ;
		p_attr[1] = (uint8_t )(p_data[0] >> 0) ;
		p_attr   += 64 ;

		// タイル
		p_dest[0] = p_data[1] ;
		p_dest[1] = p_data[2] ;
		p_dest   += 64 ;

		p_data += 3 ;
	}

	p_tcb->p_data = &p_data[0] ;
}


// -------------------------------------------------------------------------------- 
//!  背景タスク                                                                     
//!                                                                                 
//!  @param [in,out] p_tcb  背景タスク TCB                                          
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void RAMFUNC bg( BGTCB *p_tcb )
{
	int		x ;


	if( g.bg.status & BG_STATUS_STOPSCROLL ) {
		g.bg.xofs = 0 ;
		g.bg.yofs = 0 ;
		return ;
	}

	// 移動
	x         = getInt( p_tcb->x ) ;
	p_tcb->x += p_tcb->speed ;

	g.bg.x    = p_tcb->x ;
	g.bg.y    = p_tcb->y ;
	g.bg.xofs = getInt( p_tcb->x ) - x ;
	g.bg.yofs = 0 ;

	if( !g.bg.xofs ) {
		return ;
	}

	// 敵タスク生成
	createEnemy( p_tcb ) ;

	// タイル更新
	if( !(x&15) ) {
		switch( (uint8_t )p_tcb->p_data[0] ) {
			case 1:		// 繰り返し
			{
				p_tcb->p_data = (uint16_t *)&g_stgInfo[0].p_data[0] ;
				break ;
			}

			case 2:		// ボス
			{
				p_tcb->stageEvent(p_tcb, p_tcb->p_data[0]) ;
				break ;
			}

			default:
				break ;
		}

		addVblankProc(bg_vblank, p_tcb) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  背景タスクを作成                                                               
//!                                                                                 
//!  @param [in] p_parent  親タスク TCB                                             
//!  @param [in] stage     ステージ番号                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void bg_create(SYSTEMTCB *p_parent, int stage)
{
	static const BGSTAGEEVENTPROC	stageEvent[]={
										stage01_event
									} ;
	BGTCB							*p_tcb ;
	uint16_t						const *p_data ;


	DBGASSERT( sizeof( BGTCB ) <= sizeof( TCB ) ) ;

	// 画面描画
	p_data = drawScreen( &g_stgInfo[stage].p_data[0] ) ;

	// タスク生成
	p_tcb             = task_create((TCB *)p_parent, bg, TASK_PRIO_BG) ;
	p_tcb->stageEvent = stageEvent[stage] ;
	p_tcb->x          = toFp( 0 ) ;
	p_tcb->y          = toFp( 0 ) ;
	p_tcb->speed      = toFp( 0.5 ) ;
	p_tcb->p_data     = &p_data[0] ;
	p_tcb->p_enemy    = (uint16_t *)&g_stgInfo[stage].p_enemy[0] ;
	p_tcb->enemyCount = *p_tcb->p_enemy++ ;

	// ステージ作成
	p_tcb->stageEvent(p_tcb, 0) ;
}


// -------------------------------------------------------------------------------- 
//!  背景当たりチェック                                                             
//!                                                                                 
//!  @param [in] x  ｘ座標                                                          
//!  @param [in] y  ｙ座標                                                          
//!  @return        背景属性                                                        
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC bg_checkHit(int24f8_t x, int24f8_t y)
{
	int32_t		x0, y0 ;


	x0  = (g.bg.x + x) / toFp( 8 ) ;
	y0  = (g.bg.y + y) / toFp( 8 ) ;

	return g.bg.attr[y0&(VSCREEN_HEIGHT/8-1)][x0&(VSCREEN_WIDTH/8-1)] ;
}


// -------------------------------------------------------------------------------- 
//!  ｘ方向の背景当たりチェック                                                     
//!                                                                                 
//!  @param [in]  sx   開始ｘ座標                                                   
//!  @param [in]  sy   開始ｙ座標                                                   
//!  @param [in]  ex   終了ｘ座標                                                   
//!  @param [out] p_x  当たり時のｘ座標格納アドレス                                 
//!  @return           背景属性                                                     
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC bg_checkHitX(int24f8_t sx, int24f8_t sy, int24f8_t ex, int24f8_t *p_x)
{
	uint8_t		attr ;
	int32_t		x0, y0, x1 ;


	x0  =  (g.bg.x + sx) / toFp( 8 ) ;
	y0  = ((g.bg.y + sy) / toFp( 8 )) & (VSCREEN_HEIGHT / 8 - 1) ;
	x1  =  (g.bg.x + ex) / toFp( 8 ) ;

	if( x0 <= x1 ) {	// 左→右

		do {
			attr = g.bg.attr[y0][x1&(VSCREEN_WIDTH/8-1)] ;
			if( attr ) {
				*p_x = toFp( x0 * 8 - 1 ) - g.bg.x ;
				return attr ;
			}
		} while( ++x0 <= x1 ) ;

	} else {			// 左←右

		do {
			attr = g.bg.attr[y0][x1&(VSCREEN_WIDTH/8-1)] ;
			if( attr ) {
				*p_x = toFp( x0 * 8 + 8 ) - g.bg.x ;
				return attr ;
			}
		} while( --x0 >= x1 ) ;

	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ｘ方向の背景当たりチェック                                                     
//!                                                                                 
//!  @param [in]  sx   開始ｘ座標                                                   
//!  @param [in]  sy   開始ｙ座標                                                   
//!  @param [in]  ey   終了ｙ座標                                                   
//!  @param [out] p_y  当たり時のｙ座標格納アドレス                                 
//!  @return           背景属性                                                     
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC bg_checkHitY(int24f8_t sx, int24f8_t sy, int24f8_t ey, int24f8_t *p_y)
{
	uint8_t		attr ;
	int32_t		x0, y0, y1 ;


	x0 = ((g.bg.x + sx) / toFp( 8 )) & (VSCREEN_WIDTH / 8 - 1) ;
	y0 =  (g.bg.y + sy) / toFp( 8 ) ;
	y1 =  (g.bg.y + ey) / toFp( 8 ) ;

	if( y0 <= y1 ) {	// 上→下

		do {
			attr = g.bg.attr[y0&(VSCREEN_HEIGHT/8-1)][x0] ;
			if( attr ) {
				*p_y = toFp( y0 * 8 - 1 ) - g.bg.y ;
				return attr ;
			}
		} while( ++y0 <= y1 ) ;

	} else {			// 下←上

		do {
			attr = g.bg.attr[y0&(VSCREEN_HEIGHT/8-1)][sx] ;
			if( attr ) {
				*p_y = toFp( y0 * 8 + 8 ) - g.bg.y ;
				return attr ;
			}
		} while( --y0 >= y1 ) ;

	}

	return 0 ;
}
