// -------------------------------------------------------------------------------- 
//!  @file  common.c                                                                
//!  共通処理・データのソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include <math.h>
#include "hardware/structs/rosc.h"

	//! VBLANK 処理
	typedef struct {
		void						(*proc)() ;						//!< 処理
		void						*p_param ;						//!< パラメータ
	} VBLANKPROC ;


static dma_channel_config			m_dmaConfigData ;				//!< DMA チャンネルコンフィグ
static VBLANKPROC					m_vblankProc[VBLANKPROC_MAX] ;	//!< VBlank 時処理
static uint8_t						m_vblankProcCount ;				//!< VBlank 時処理残り数

GLOBALVAR							g ;								//!< グローバル変数

//! ヴィクセンデータ転送データ
const uint8_t						g_xferVixen[4][2]={
										{ XFERDATA_VIXEN0_OFFSET, XFERDATA_VIXEN0_COUNT },
										{ XFERDATA_VIXEN1_OFFSET, XFERDATA_VIXEN1_COUNT },
										{ XFERDATA_VIXEN2_OFFSET, XFERDATA_VIXEN2_COUNT },
										{ XFERDATA_VIXEN3_OFFSET, XFERDATA_VIXEN3_COUNT }
									} ;


// -------------------------------------------------------------------------------- 
//!  VSync 割り込みサービスルーチン                                                 
//!                                                                                 
//!  @param [in] gpio    GPIO 番号                                                  
//!  @param [in] events  イベント                                                   
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
static void RAMFUNC isr_vsync(uint gpio, uint32_t events)
{
	sound_vsync() ;

	if( g.vsyncProc != NULL ) {
		g.vsyncProc( g.p_vsyncParam ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void common_init( void )
{
	int		angle, x ,y ;

	// DMA 初期化
	dma_channel_claim( DMA_CH ) ;
	m_dmaConfigData = dma_channel_get_default_config( DMA_CH ) ;
	channel_config_set_write_increment(&m_dmaConfigData, true) ;

	// 初期化
	setMemory(&g, 0, sizeof( g )) ;
	g.highScore       = 500 ;
	m_vblankProcCount = VBLANKPROC_MAX ;

	// サインデータ作成
	for(angle=0; angle<256; angle++) {
		g.sinData[angle] = (int16_t )(sin( M_PI / (256 / 2) * angle ) * 256.) ;
	}

	// 角度データ作成
	for(y=0; y<256/16; y++) {
		for(x=0; x<512/16; x++) {
			g.angleData[y][x] = (uint8_t )(atan2(y*16, x*16) / M_PI * (256. / 2.)) ;
		}
	}

	vga_init() ;
	sound_init() ;
	task_init() ;

	g.padId = pad_init() ;

	gpio_set_irq_enabled_with_callback(VGA_CFG_PIN_VSYNC, GPIO_IRQ_EDGE_FALL, true, isr_vsync) ;

	vga_start() ;

	g.cfg.button[0] = 0 ;
	g.cfg.button[1] = 1 ;
	g.cfg.button[2] = 1 + (g.padId == PAD_ID_MD) ;
	g.cfg.rapid     = 1 ;
}


// -------------------------------------------------------------------------------- 
//!  メモリ設定                                                                     
//!                                                                                 
//!  @param [out]  p_dest  メモリアドレス                                           
//!  @param [in]   value   値                                                       
//!  @param [in]   size    サイズ                                                   
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void RAMFUNC setMemory(void *p_dest, uint8_t value, int size)
{
	uint8_t		*p_dest8, *p_dest8End ;
	uint32_t	value32 ;


	DBGASSERT( size ) ;

	p_dest8 = p_dest ;

	// 16 バイト以下は CPU でコピー
	if( size < 32 ) {
		p_dest8End = &p_dest8[size] ;
		while( p_dest8 != p_dest8End ) {
			*p_dest8++ = value ;
		}
		return ;
	}

	value32  = ((uint32_t )value << 8) | value ;
	value32 |= (value32 << 16) ;

	// 先頭のアラインメントを揃える
	if( (uintptr_t )p_dest8 & 1 ) {
		*p_dest8++ = (uint8_t )value32 ;
		size-- ;
	}

	if( (uintptr_t )p_dest8 & 2 ) {
		*(uint16_t *)p_dest8 = (uint16_t )value32 ;
		p_dest8             += 2 ;
		size                -= 2 ;
	}

	// DMA 転送
	dma_channel_wait_for_finish_blocking( DMA_CH ) ;
	channel_config_set_transfer_data_size(&m_dmaConfigData, DMA_SIZE_32) ;
	channel_config_set_read_increment(&m_dmaConfigData, false) ;
	dma_channel_configure(DMA_CH, &m_dmaConfigData, &p_dest8[0], &value32, size/4, true) ;

	p_dest8 += (size & ~3) ;

	// 端数
	if( size & 2 ) {
		*(uint16_t *)p_dest8 = (uint16_t )value32 ;
		p_dest8             += 2 ;
	}

	if( size & 1 ) {
		*p_dest8 = (uint8_t )value32 ;
	}

	dma_channel_wait_for_finish_blocking( DMA_CH ) ;
}


// -------------------------------------------------------------------------------- 
//!  32bit 単位のメモリ設定                                                         
//!                                                                                 
//!  @param [out]  p_dest  メモリアドレス                                           
//!  @param [in]   value   値                                                       
//!  @param [in]   count   設定数                                                   
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void RAMFUNC setMemory32(void *p_dest, uint32_t value, int count)
{
	DBGASSERT( count ) ;

	dma_channel_wait_for_finish_blocking( DMA_CH ) ;
	channel_config_set_transfer_data_size(&m_dmaConfigData, DMA_SIZE_32) ;
	channel_config_set_read_increment(&m_dmaConfigData, false) ;
	dma_channel_configure(DMA_CH, &m_dmaConfigData, p_dest, &value, count, true) ;
	dma_channel_wait_for_finish_blocking( DMA_CH ) ;
}


// -------------------------------------------------------------------------------- 
//!  乱数を取得                                                                     
//!                                                                                 
//!  @return 乱数                                                                   
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC getRandom( void )
{
	uint8_t		value ;


	value  = ((rosc_hw->randombit << 31) >> 31) ;
	value |= ((rosc_hw->randombit << 31) >> 30) ;
	value |= ((rosc_hw->randombit << 31) >> 29) ;
	value |= ((rosc_hw->randombit << 31) >> 28) ;
	value |= ((rosc_hw->randombit << 31) >> 27) ;
	value |= ((rosc_hw->randombit << 31) >> 26) ;
	value |= ((rosc_hw->randombit << 31) >> 25) ;
	value |= ((rosc_hw->randombit << 31) >> 24) ;

	return value ;
}


// -------------------------------------------------------------------------------- 
//!  角度を取得                                                                     
//!                                                                                 
//!  @param [in] sx  開始ｘ座標                                                     
//!  @param [in] sy  開始ｙ座標                                                     
//!  @param [in] ex  終了ｘ座標                                                     
//!  @param [in] ey  終了ｙ座標                                                     
//!  @return         角度                                                           
// -------------------------------------------------------------------------------- 
uint8_t RAMFUNC getAngle(int sx, int sy, int ex, int ey)
{
	uint8_t		angle ;
	int			x, y ;


	x = ex - sx ;
	y = ey - sy ;

	if( x >= 0 ) {
		if( y >= 0 ) {		// 右下
			angle =       g.angleData[( y+8)/16][( x+8)/16] ;
		} else {			// 右上
			angle =   0 - g.angleData[(-y+8)/16][( x+8)/16] ;
		}
	} else {
		if( y >= 0 ) {		// 左下
			angle = 128 - g.angleData[( y+8)/16][(-x+8)/16] ;
		} else {			// 左上
			angle = 128 + g.angleData[(-y+8)/16][(-x+8)/16] ;
		}
	}

	return angle ;
}


// -------------------------------------------------------------------------------- 
//!  V-BLANK 用処理を追加                                                           
//!                                                                                 
//!  @param [in] proc    処理                                                       
//!  @param [in] p_param パラメータ                                                 
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
void RAMFUNC addVblankProc(void (*proc)(), void *p_param)
{
	VBLANKPROC	*p_vblankProc ;


	DBGASSERT( m_vblankProcCount ) ;

	p_vblankProc          = &m_vblankProc[--m_vblankProcCount] ;
	p_vblankProc->proc    = proc ;
	p_vblankProc->p_param = p_param ;
}


// -------------------------------------------------------------------------------- 
//!  VBlank 用処理を実行                                                            
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC execVblankProc( void )
{
	VBLANKPROC	*p_vblankProc, *p_vblankProcEnd ;


	p_vblankProc      = &m_vblankProc[m_vblankProcCount] ;
	p_vblankProcEnd   = &m_vblankProc[VBLANKPROC_MAX] ;
	m_vblankProcCount = VBLANKPROC_MAX ;

	while( p_vblankProc != p_vblankProcEnd ) {
		p_vblankProc->proc( p_vblankProc->p_param ) ;
		p_vblankProc++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  データ転送                                                                     
//!  ROM から RAM へのデータ転送を行うため、VGA 出力に影響の出る DMA は使用しない。 
//!  strm/ldrmを使用した高速なメモリコピー関数を用意し、その関数を呼ぶようにする。  
//!                                                                                 
//!  @param [in]  index  転送データ開始インデックス                                 
//!  @param [out] count  転送データ数                                               
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
void RAMFUNC xferData(uint32_t index, uint32_t count)
{
	const XFERDATA		*p_xferData, *p_xferDataEnd ;


	p_xferData    = &g_xferData[index] ;
	p_xferDataEnd = &p_xferData[count] ;

	while( 1 ) {
		memcpy(p_xferData->p_dest, p_xferData->p_srce, (size_t )p_xferData->count*4) ;

		if( ++p_xferData == p_xferDataEnd ) {
			break ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  タイル面へ文字列を書き込む（8x8）                                              
//!                                                                                 
//!  @param [in] x         ｘ座標                                                   
//!  @param [in] y         ｙ座標                                                   
//!  @param [in] p_string  文字列                                                   
//!  @param [in] palette   パレット番号                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void RAMFUNC writeTileString8(int x, int y, char const *p_string, uint8_t palette)
{
	uint8_t		code ;
	uint16_t	tile, *p_tile ;


	p_tile = &g.vram.tile[y][x] ;
	tile   = palette << VGATILE_PC_BIT ;

	while( 1 ) {
		code = *p_string++ ;
		if( code == '\0' ) {
			break ;
		}

		*p_tile++ = tile | code ;
	}
}


// -------------------------------------------------------------------------------- 
//!  タイル面へ文字列を書き込む（16x16）                                            
//!                                                                                 
//!  @param [in] x         ｘ座標                                                   
//!  @param [in] y         ｙ座標                                                   
//!  @param [in] p_string  文字列                                                   
//!  @param [in] palette   パレット番号                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void RAMFUNC writeTileString16(int x, int y, char const *p_string, uint8_t palette)
{
	uint8_t		code ;
	uint16_t	tile, *p_tile ;


	p_tile = &g.vram.tile[y][x] ;
	tile   = palette << VGATILE_PC_BIT ;

	while( 1 ) {
		code = *p_string++ ;
		if( code == '\0' ) {
			break ;
		}

		if( code != ' ' ) {
			code       = 0x60 + ((code - 0x41) * 4) ;
			p_tile[0]  = tile | (code + 0) ;
			p_tile[1]  = tile | (code + 2) ;
			p_tile[64] = tile | (code + 1) ;
			p_tile[65] = tile | (code + 3) ;
		} else {
			p_tile[0]  = 0 ;
			p_tile[1]  = 0 ;
			p_tile[64] = 0 ;
			p_tile[65] = 0 ;
		}
		p_tile += 2 ;
	}
}
