// -------------------------------------------------------------------------------- 
//!  @file  common.h                                                                
//!  共通処理・データのヘッダファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#ifndef __COMMON_H__INCLUDED__
#define __COMMON_H__INCLUDED__


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pico/stdlib.h>
#include "config.h"
#include "libvga/vga/vga.h"
#include "libvga/pad/pad.h"

#ifdef  DEBUG
#define DBGASSERT( e )				if( !(e) ) { printf("******** %s(%d)\n", __func__, __LINE__) ;  while( 1 ) {} }
																	//!< デバッグ用アサート
#define DBGTRACE(fmt, ...)			printf(fmt, __VA_ARGS__)		//!< デバッグ用トレース
#else
#define DBGASSERT( e )												//!< デバッグ用アサート
#define DBGTRACE(fmt, ...)											//!< デバッグ用トレース
#endif//DEBUG

#define MEMBERSTRUCT(_s, _m, _p)	((_s *)((uintptr_t )(_p) - (uintptr_t )&((_s *)0)->_m))
																	//!< メンバーアドレスから構造体の先頭アドレスを取得

#define __enable_irq()				__asm volatile ("cpsie i")		//!< 割り込み許可
#define __disable_irq()				__asm volatile ("cpsid i")		//!< 割り込み禁止

typedef int16_t						int8f8_t ;						//!< 固定小数点（整数部  8bit、小数部 8bit）
typedef int32_t						int24f8_t ;						//!< 固定小数点（整数部 24bit、小数部 8bit）
typedef void (*VSYNCPROC)() ;										//!< VSync 処理


	enum {
		DMA_CH						=	2,							//!< DMA チャンネル
		FP_ONE_						=	8,							//!< 固定小数点１ビット数
		FP_ONE						=	1 << FP_ONE_,				//!< 固定小数点１
		SCREEN_WIDTH				=	320,						//!< 画面幅
		SCREEN_HEIGHT				=	224,						//!< 画面高さ
		VSCREEN_WIDTH				=	512,						//!< 仮想画面幅
		VSCREEN_HEIGHT				=	256,						//!< 仮想画面高さ
		TASK_MAX					=	128,						//!< タスク最大数
		TASK_WORK					=	64,							//!< タスク作業領域サイズ
		PLAYERTASK_MAX				=	61,							//!< プレイヤータスク最大数
		ENEMYTASK_MAX				=	32,							//!< 敵タスク最大数
		VBLANKPROC_MAX				=	16,							//!< VBlank 時処理最大数
		VIXEN_POS_COUNT				=	64,							//!< 位置数
		VIXEN_POS_MASK				=	VIXEN_POS_COUNT - 1			//!< 位置数マスク
	} ;

	//! タスク優先度
	enum {
		TASK_PRIO_SYSTEM			=	0,							//!< システムタスク
		TASK_PRIO_BG				=	16,							//!< 背景タスク
		TASK_PRIO_PLAYER			=	32,							//!< プレイヤータスク
		TASK_PRIO_MULTIPLE			=	TASK_PRIO_PLAYER+1,			//!< マルチプルタスク
		TASK_PRIO_SHIELD			=	TASK_PRIO_PLAYER+2,			//!< シールドタスク
		TASK_PRIO_WEAPON			=	TASK_PRIO_PLAYER+3,			//!< プレイヤー武器タスク
		TASK_PRIO_ENEMY0			=	128,						//!< 敵タスク
		TASK_PRIO_ENEMY1			=	192,						//!< 敵タスク（オブジェクトなし）
		TASK_PRIO_POST				=	255							//!< 後処理タスク
	} ;

	//! オブジェクト優先度
	enum {
		OBJ_PRIO0					=	3 << 14,					//!< 高優先度
		OBJ_PRIO1					=	2 << 14,					//!<    ↑   
		OBJ_PRIO2					=	1 << 14,					//!<    ↓   
		OBJ_PRIO3					=	0 << 14						//!< 低優先度
	} ;

	//! パワーメーター状態
	enum {
		POWERMETER_STATUS_SPEED		=	1 << 1,						//!< スピード
		POWERMETER_STATUS_MISSILE	=	1 << 2,						//!< ミサイル
		POWERMETER_STATUS_DOUBLE	=	1 << 3,						//!< ダブル
		POWERMETER_STATUS_LASER		=	1 << 4,						//!< レーザー
		POWERMETER_STATUS_MULTIPLE	=	1 << 5,						//!< マルチプル
		POWERMETER_STATUS_SHIELD	=	1 << 6						//!< シールド
	} ;

	//! 背景属性
	enum {
		BG_ATTR_WALL				=	1 << 7,						//!< 壁
	} ;

	//! 背景状態
	enum {
		BG_STATUS_STOPSCROLL		=	1 << 0,						// スクロール停止
	} ;

	//! 転送データ
	typedef struct {
		void						const *p_srce ;					//!< 転送元
		void						*p_dest ;						//!< 転送先
		uint16_t					count ;							//!< 転送数
	} XFERDATA ;

	//! パターンデータ
	typedef struct {
		int8_t						x ;								//!< ｘ座標オフセット
		int8_t						y ;								//!< ｙ座標オフセット
		uint16_t					attr ;							//!< 属性
	} PATDATA ;

	//! パターン情報
	typedef struct {
		uint16_t					index ;							//!< パターンデータインデックス
		uint8_t						count ;							//!< パターンデータ数
		uint8_t						flag ;							//!< フラグ
		int8_t						x ;								//!< ｘ座標オフセット
		int8_t						y ;								//!< ｙ座標オフセット
		uint8_t						w ;								//!< 幅
		uint8_t						h ;								//!< 高さ
	} PATINFO ;

	//! ステージ情報
	typedef struct {
		uint16_t					const *p_data ;					//!< 背景データ
		uint32_t					dataSize ;						//!< データサイズ
		uint16_t					const *p_enemy ;				//!< 敵データ
		uint32_t					enemySize ;						//!< 敵データサイズ
	} STGINFO ;

	//! グローバル変数
	typedef struct {
		int16_t						sinData[256] ;					//!< サインデータ
		uint8_t						angleData[16][32] ;				//!< 角度データ
		VSYNCPROC					vsyncProc ;						//!< VSync 処理
		void						*p_vsyncParam ;					//!< Vsync 処理パラメータ
		uint8_t						requestSE[DPCM3_CFG_CHANNEL_COUNT] ;
																	//!< SE リクエスト
		uint8_t						padId ;							//!< パッド ID
		uint8_t						task ;							//!< タスク処理フラグ
		uint16_t					count1 ;						//!< １フレームカウンタ
		uint16_t					count2 ;						//!< ２フレームカウンタ
		uint16_t					count3 ;						//!< ３フレームカウンタ
		uint16_t					count4 ;						//!< ４フレームカウンタ
		uint16_t					count8 ;						//!< ８フレームカウンタ
		uint32_t					highScore ;						//!< 最高得点
		uint32_t					score ;							//!< 得点
		uint8_t						powerCapsule ;					//!< 取得パワーカプセル数

		//! VRAM
		struct {
			uint32_t				pat[2048][32/4] ;				//!< パターンデータ
			uint16_t				pal[32][16] ;					//!< パレットデータ
			uint16_t				tile[VGA_TILE_HEIGHT_COUNT][VGA_TILE_WIDTH_COUNT] ;
																	//!< タイルデータ
			VGASPR					spr[2][VGA_SPR_MAX] ;			//!< スプライトデータ
			uint16_t				txt[2][VGA_TXT_WIDTH_COUNT] ;	//!< テキストデータ
			uint8_t					page ;							//!< 表示ページ
		} vram ;

		//! 構成
		struct {
			uint8_t					button[3] ;						//!< ボタン位置
			uint8_t					rapid ;							//!< 連射
			uint8_t					player ;						//!< プレイヤー数
		} cfg ;

		//! パッド状態
		struct {
			uint16_t				norm ;							//!< 通常
			uint16_t				old ;							//!< 以前
			uint16_t				once ;							//!< １度のみ
			uint16_t				conv ;							//!< 変換
			uint16_t				rapid ;							//!< 変換＋連射
		} pad ;

		//! パワーメーター
		struct {
			uint8_t					status ;						//!< パワーメーター状態
			uint8_t					playerType ;					//!< プレーヤータイプ番号
			uint8_t					multiple ;						//!< マルチプル番号
			uint8_t					shield ;						//!< シールド番号
		} powerMeter ;

		//! ステージ
		struct {
			uint8_t					no ;							//!< ステージ番号
			uint32_t				count0 ;						//!< 汎用カウンタ０
		} stage ;

		//! プレイヤー
		struct {
			int24f8_t				x ;								//!< プレイヤーｘ座標
			int24f8_t				y ;								//!< プレイヤーｙ座標
			uint8_t					remain ;						//!< 残機数
		} player ;

		//! ヴィクセン
		struct {
			int16_t					pos[VIXEN_POS_COUNT][2] ;		//!< 位置
			uint8_t					multipleCount ;					//!< マルチプル数
		} vixen ;

		//! 背景
		struct {
			uint16_t				status ;						//!< 状態
			int24f8_t				x ;								//!< 背景ｘ座標
			int24f8_t				y ;								//!< 背景ｙ座標
			int16_t					xofs ;							//!< 背景移動ｘ座標オフセット
			int16_t					yofs ;							//!< 背景移動ｙ座標オフセット
			uint8_t					attr[VSCREEN_HEIGHT/8][VSCREEN_WIDTH/8] ;
																	//!< 背景属性データ
		} bg ;

		//! 敵
		struct {
			uint8_t					remain ;						//!< 残機数
		} enemy ;
	} GLOBALVAR ;


extern GLOBALVAR					g ;								//!< グローバル変数

extern const uint8_t				g_xferVixen[4][2] ;				//!< ヴィクセンデータ転送データ


#include "data.h"
#include "task.h"
#include "sound.h"
#include "sound.h"


extern void            common_init( void ) ;
extern void    RAMFUNC setMemory(void *p_dest, uint8_t value, int size) ;
extern void    RAMFUNC setMemory32(void *p_dest, uint32_t value, int count) ;
extern uint8_t RAMFUNC getRandom( void ) ;
extern uint8_t RAMFUNC getAngle(int sx, int sy, int ex, int ey) ;
extern void    RAMFUNC addVblankProc(void (*proc)(), void *p_param) ;
extern void    RAMFUNC execVblankProc( void ) ;
extern void    RAMFUNC xferData(uint32_t index, uint32_t count) ;
extern void    RAMFUNC writeTileString8(int x, int y, char const *p_string, uint8_t palette) ;
extern void    RAMFUNC writeTileString16(int x, int y, char const *p_string, uint8_t palette) ;


// -------------------------------------------------------------------------------- 
//!  数値から固定小数点へ変換                                                       
//!                                                                                 
//!  @param [in] n  数値                                                            
//!  @return        固定小数点                                                      
// -------------------------------------------------------------------------------- 
#define toFp( n )					(FP_ONE*(n))


// -------------------------------------------------------------------------------- 
//!  固定小数点の整数部を取得                                                       
//!                                                                                 
//!  @param [in] fp  固定小数点                                                     
//!  @return         整数部                                                         
// -------------------------------------------------------------------------------- 
static inline int32_t getInt( int24f8_t fp )
{
	return (fp/FP_ONE) ;
}


// -------------------------------------------------------------------------------- 
//!  サインを取得                                                                   
//!                                                                                 
//!  @param [in] angle  角度                                                        
//!  @return            固定小数のサイン値                                          
// -------------------------------------------------------------------------------- 
static inline int24f8_t getSin( uint8_t angle )
{
	return (int24f8_t )g.sinData[angle] ;
}


// -------------------------------------------------------------------------------- 
//!  コサインを取得                                                                 
//!                                                                                 
//!  @param [in] angle  角度                                                        
//!  @return            固定小数のコサイン値                                        
// -------------------------------------------------------------------------------- 
static inline int24f8_t getCos( uint8_t angle )
{
	return (int24f8_t )g.sinData[(uint8_t )(angle+64)] ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 割り込み処理を設定                                                       
//!                                                                                 
//!  @param [in] vsyncProc  VSync 割り込み処理（NULL 時は解除）                     
//!  @param [in] p_param    パラメータ                                              
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static inline void setVSyncProc(VSYNCPROC vsyncProc, void *p_param)
{
	g.vsyncProc    = vsyncProc ;
	g.p_vsyncParam = vsyncProc ;
}


#endif//__COMMON_H__INCLUDED__
