// -------------------------------------------------------------------------------- 
//!  @file  hit.c                                                                   
//!  当たり管理のソースファイル                                                     
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"


	enum {
		ENEMYQUEUE_WIDTH			=	32,							//!< 敵キュー幅
		ENEMYQUEUE_COUNT			=	SCREEN_WIDTH / ENEMYQUEUE_WIDTH,
																	//!< 敵キュー最大数
	} ;

	//! プレイヤーキュー（最後から追加）
	typedef struct {
		HIT							*p_hit[PLAYERTASK_MAX] ;		//!< 当たり情報アドレス
		HIT							**pp_hitEnd ;					//!< 当たり情報アドレスの終端
	} PLAYERQUEUE ;

	//! 敵キュー（先頭から追加）
	typedef struct {
		HIT							*p_hit[ENEMYTASK_MAX] ;			//!< 当たり情報アドレス
		HIT							**pp_hitEnd ;					//!< 当たり情報アドレスの終端
	} ENEMYQUEUE ;


static PLAYERQUEUE					m_playerQueue ;					//!< プレイヤーキュー
static ENEMYQUEUE					m_enemyQueue[ENEMYQUEUE_COUNT] ;//!< 敵キュー


// -------------------------------------------------------------------------------- 
//!  敵当たり処理                                                                   
//!                                                                                 
//!  @param [in] p_playerHit  プレイヤー当たり情報                                  
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
static inline void hitEnemy( HIT *p_player )
{
	ENEMYQUEUE	*p_enemyQueue, *p_enemyQueueEnd ;
	HIT			*p_enemy, **pp_hit, **pp_hitEnd ;
	uint32_t	x, y ;


	x               = (uint32_t )p_player->x + p_player->w ;
	y               = (uint32_t )p_player->y + p_player->h ;
	p_enemyQueue    = &m_enemyQueue[p_player->x/ENEMYQUEUE_WIDTH] ;
	p_enemyQueueEnd = &m_enemyQueue[x          /ENEMYQUEUE_WIDTH] ;

	do {
		pp_hit    = &p_enemyQueue->p_hit[0] ;
		pp_hitEnd = p_enemyQueue->pp_hitEnd ;

		while( pp_hit != pp_hitEnd ) {
			p_enemy = *pp_hit++ ;

			do {
				if( !(p_player->flag&p_enemy->flag) ) {
					break ;
				}

				if( (uint32_t )(x-p_enemy->x-1) >= (uint32_t )(p_player->w+p_enemy->w-1) ) {
					break ;
				}

				if( (uint32_t )(y-p_enemy->y-1) >= (uint32_t )(p_player->h+p_enemy->h-1) ) {
					break ;
				}

				p_enemy->proc(p_enemy, p_player) ;

				if( !p_player->flag ) {
					return ;
				}
			} while( 0 ) ;
		}
	} while( ++p_enemyQueue < p_enemyQueueEnd ) ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC hit_init( void )
{
	ENEMYQUEUE	*p_enemyQueue, *p_enemyQueueEnd ;

	// プレイヤーキュー初期化
	m_playerQueue.pp_hitEnd = &m_playerQueue.p_hit[PLAYERTASK_MAX] ;

	// 敵キュー初期化
	p_enemyQueue    = &m_enemyQueue[0] ;
	p_enemyQueueEnd = &m_enemyQueue[ENEMYQUEUE_COUNT] ;

	do {
		p_enemyQueue[0].pp_hitEnd = &p_enemyQueue[0].p_hit[0] ;
		p_enemyQueue[1].pp_hitEnd = &p_enemyQueue[1].p_hit[0] ;
		p_enemyQueue             += 2 ;
	} while( p_enemyQueue != p_enemyQueueEnd ) ;
}


// -------------------------------------------------------------------------------- 
//!  当たり情報を設定                                                               
//!                                                                                 
//!  @param [in,out] p_obj  オブジェクトアドレス                                    
//!  @retval         true   当たり対象範囲内                                        
//!  @retval         false  当たり対象範囲外                                        
// -------------------------------------------------------------------------------- 
bool RAMFUNC hit_set( OBJ *p_obj )
{
	PATINFO		const *p_patInfo ;
	HIT			*p_hit ;
	int			x0, x1, y0, y1 ;


	p_patInfo   = &g_patInfo[p_obj->pattern] ;
	p_hit       = &p_obj->hit ;
	p_hit->flag = 0 ;

	if( !p_patInfo->flag ) {
		return false ;
	}

	// ｘ方向
	x0 = getInt( p_obj->x ) + p_patInfo->x ;
	if( x0 >= SCREEN_WIDTH ) {
		return false ;
	}

	x1 = x0 + p_patInfo->w ;
	if( x1 < 0 ) {
		return false ;
	}

	if( x0 < 0 ) {
		x0 = 0 ;
	} else
	if( x1 >= (SCREEN_WIDTH-1) ) {
		x1  = (SCREEN_WIDTH-1) ;
	}

	p_hit->x = (uint16_t )x0 ;
	p_hit->w = (uint8_t  )(x1 - x0) ;

	if( !p_hit->w ) {
		return false ;
	}

	// ｙ方向
	y0 = getInt( p_obj->y ) + p_patInfo->y ;
	if( y0 >= SCREEN_HEIGHT ) {
		return false ;
	}

	y1 = y0 + p_patInfo->h ;
	if( y1 < 0 ) {
		return false ;
	}

	if( y0 < 0 ) {
		y0 = 0 ;
	} else
	if( y1 >= (SCREEN_HEIGHT-1) ) {
		y1  = (SCREEN_HEIGHT-1) ;
	}

	p_hit->y = (uint8_t )y0 ;
	p_hit->h = (uint8_t )(y1 - y0) ;

	if( !p_hit->h ) {
		return false ;
	}

	p_hit->flag = p_patInfo->flag ;

	hit_add( p_hit ) ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  当たり情報キューへ追加                                                         
//!                                                                                 
//!  @param [in,out] p_hit  当たり情報                                              
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
void RAMFUNC hit_add( HIT *p_hit )
{
	ENEMYQUEUE	*p_enemyQueue, *p_enemyQueueEnd ;


	if( !(p_hit->flag&HIT_FLAG_ENEMY) ) {	// プレイヤー

		*(--m_playerQueue.pp_hitEnd) = p_hit ;

	} else {								// 敵

		p_enemyQueue    = &m_enemyQueue[            p_hit->x          /ENEMYQUEUE_WIDTH] ;
		p_enemyQueueEnd = &m_enemyQueue[((uint32_t )p_hit->x+p_hit->w)/ENEMYQUEUE_WIDTH] ;

		do {
			*p_enemyQueue->pp_hitEnd++ = p_hit ;
		} while( ++p_enemyQueue < p_enemyQueueEnd ) ;

	}
}


// -------------------------------------------------------------------------------- 
//!  当たり計算                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC hit_calc( void )
{
	HIT		**pp_hit ;


	pp_hit = m_playerQueue.pp_hitEnd ;

	while( pp_hit != &m_playerQueue.p_hit[PLAYERTASK_MAX] ) {
		hitEnemy( *pp_hit++ ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  当たり処理をしない                                                             
//!                                                                                 
//!  @param [in,out] p_self   自分                                                  
//!  @param [in,out] p_rival  相手                                                  
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
void RAMFUNC hit_nothing(HIT *p_self, HIT *p_rival)
{
}
