// -------------------------------------------------------------------------------- 
//!  @file  obj.c                                                                   
//!  オブジェクト管理のソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"


static OBJ							*m_p_objList=NULL ;				//!< オブジェクトリスト


// -------------------------------------------------------------------------------- 
//!  オブジェクトリストへ追加                                                       
//!                                                                                 
//!  @param [in] p_obj  オブジェクトアドレス                                        
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
void RAMFUNC addObjList( OBJ *p_obj )
{
	OBJ			**pp_prev, *p_next ;
	uint16_t	prio ;


	prio = p_obj->prio ;

	pp_prev = &m_p_objList ;
	while( 1 ) {
		p_next = *pp_prev ;
		if( p_next == NULL ) {
			break ;
		}

		if( prio >= p_next->prio ) {
			break ;
		}

		pp_prev = &p_next->p_next ;
	}

	p_obj->p_next = p_next ;
	*pp_prev      = p_obj ;
}


// -------------------------------------------------------------------------------- 
//!  パターンを設定                                                                 
//!                                                                                 
//!  @param [in,out] p_obj    オブジェクトアドレス                                  
//!  @param [in]     pattern  パターン番号                                          
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_setPattern(OBJ *p_obj, uint16_t pattern)
{
	p_obj->pattern = pattern ;

	if( !pattern || !g.task ) {
		return ;
	}

	hit_set( p_obj ) ;

	addObjList( p_obj ) ;
}


// -------------------------------------------------------------------------------- 
//!  アクションを設定                                                               
//!                                                                                 
//!  @param [in,out] p_obj   オブジェクトアドレス                                   
//!  @param [in]     action  アクション番号                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_setAction(OBJ *p_obj, uint8_t action)
{
	p_obj->status  &= ~OBJ_STATUS_ACTEND ;
	p_obj->action   = action ;
	p_obj->p_action = &g_actData[action][0] ;
	p_obj->pattern  = p_obj->p_action[0] ;
	p_obj->count    = p_obj->p_action[1] ;

	if( !p_obj->pattern || !g.task ) {
		return ;
	}

	hit_set( p_obj ) ;

	addObjList( p_obj ) ;
}


// -------------------------------------------------------------------------------- 
//!  アクションを更新                                                               
//!                                                                                 
//!  @param [in,out] p_obj  オブジェクトアドレス                                    
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_updateAction( OBJ *p_obj )
{
	const uint16_t	*p_action ;
	uint16_t		pattern, count ;


	do {
		// アクション終了
		if( p_obj->status & OBJ_STATUS_ACTEND ) {
			break ;
		}

		if( --p_obj->count ) {
			break ;
		}

		p_action = &p_obj->p_action[2] ;
		pattern  = p_action[0] ;
		count    = p_action[1] ;

		if( !count ) {

			// 終了
			if( !pattern ) {
				p_obj->status |= OBJ_STATUS_ACTEND ;
				break ;
			}

			// ループ
			p_action -= (pattern * 2) ;
			pattern   = p_action[0] ;
			count     = p_action[1] ;
		}

		p_obj->p_action = &p_action[0] ;
		p_obj->pattern  = pattern ;
		p_obj->count    = count ;
	} while( 0 ) ;

	obj_setPattern(p_obj, p_obj->pattern) ;
}


// -------------------------------------------------------------------------------- 
//!  スプライト面 VRAM データを作成                                                 
//!                                                                                 
//!  @param [out] p_spr  スプライト面 VRAM データ                                   
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_createSprite( VGASPR *p_spr )
{
	VGASPR		*p_sprEnd ;
	OBJ			*p_obj ;
	PATINFO		const *p_patInfo ;
	PATDATA		const *p_patData, *p_patDataEnd ;
	uint8_t		lineCount[1+(SCREEN_HEIGHT/16)+1] ;
	int16_t		bx, by, sx, sy ;
	uint32_t	attr ;
	bool		miss ;
	int			index ;

	// 初期化
	setMemory(&lineCount[0], 0, sizeof( lineCount )) ;
	setMemory32(&p_spr[0], 0x01000000, (sizeof( VGASPR )*VGA_SPR_MAX)/4) ;

	// スプライト設定
	p_sprEnd    = &p_spr[VGA_SPR_MAX] ;
	p_obj       = m_p_objList ;
	m_p_objList = NULL ;

	while( p_obj != NULL ) {
		p_patInfo    = &g_patInfo[p_obj->pattern] ;
		p_patData    = &g_patData[p_patInfo->index] ;
		p_patDataEnd = &p_patData[p_patInfo->count] ;
		bx           = getInt( p_obj->x ) ;
		by           = getInt( p_obj->y ) ;
		attr         = (uint32_t )((p_obj->status >> OBJ_STATUS_SPRPRIO_) ^ 1) << VGASPR_ATTR_PR_BIT ;
		miss         = false ;

		do {
			// スプライト数オーバー
			if( p_spr == p_sprEnd ) {
				miss = true ;
				break ;
			}

			// ｘ範囲
			sx = bx + p_patData->x ;
			if( (uint32_t )(sx+15) < (uint32_t )(15+SCREEN_WIDTH-1) ) {

				// ｙ範囲
				sy = by + p_patData->y ;
				if( (uint32_t )(sy+15) < (15+SCREEN_HEIGHT-1) ) {

					// 水平表示スプライト数オーバー
					index = sy / 16 ;
					if( lineCount[index] == 16 ) {
						miss = true ;
						break ;
					}
					lineCount[index]++ ;

					p_spr->attr = attr | p_patData->attr ;
					p_spr->x    = sx ;
					p_spr->y    = sy ;
					p_spr++ ;
				}
			}
		} while( ++p_patData != p_patDataEnd ) ;

		// 優先度調整
		if( miss == false ) {
			p_obj->prio &= 0xc00f ;
			if( p_obj->prio & 0x3ff0 ) {
				p_obj->prio -= 0x0010 ;
			}
		} else {
			p_obj->prio += 0x0180 ;
		}

		p_obj = p_obj->p_next ;
	}
}
