// -------------------------------------------------------------------------------- 
//!  @file  sound.c                                                                 
//!  サウンド管理のソースファイル                                                   
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "libvga/dac/mcp49x1.h"
#include "libvga/pcm/dpcm3.h"
#include "libvga/pcm/scc.h"


#ifdef  SOUNDDAC_CFG_DAC_MCP49x1
#define DAC_INIT					mcp49x1_init					//!< DAC 初期化処理
#define DAC_WRITE					mcp49x1_write					//!< DAC 書き込み処理
#define DAC_WAIT					mcp49x1_wait					//!< DAC 書き込み待ち処理
#endif//SOUNDDAC_CFG_DAC_MCP49x1


static struct repeating_timer		m_timer ;						//!< 繰り返しタイマー
static DPCM3INFO					m_dpcm3Info ;					//!< 3bit DPCM 情報
static PCMSCCINFO					m_sccInfo ;						//!< SCC 情報
static uint16_t						m_output ;						//!< 出力値
static uint16_t						const *m_p_bgmHead ;			//!< BGM データ先頭
static uint16_t						const *m_p_bgm ;				//!< BGM データ
static uint16_t						m_volume ;						//!< フェードアウトボリューム


// -------------------------------------------------------------------------------- 
//!  繰り返しタイマー割り込みサービスルーチン                                       
//!                                                                                 
//!  @param [in] p_timer  繰り返しタイマー                                          
//!  @retval     true     繰り返し                                                  
//!  @retval     false    停止                                                      
// -------------------------------------------------------------------------------- 
static bool RAMFUNC isr_timer( struct repeating_timer *p_timer )
{
	int32_t		output ;


	DAC_WRITE( m_output ) ;

	output   = dpcm3_update( &m_dpcm3Info ) ;
	output  += pcm_scc_update( &m_sccInfo ) ;
	m_output = (uint16_t )(output + 0x0800) ;

	DAC_WAIT() ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  SCC 更新                                                                       
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static inline void updateSCC( void )
{
	uint16_t	data[5] ;
	uint16_t	const *p_bgm ;
	int			index, max, volume ;


	if( m_p_bgmHead == NULL ) {
		return ;
	}

	// 停止・繰り返し
	p_bgm = &m_p_bgm[0] ;
	if( *p_bgm >= 0xfffe ) {

		// 停止
		if( *p_bgm == 0xfffe ) {
			m_p_bgmHead = NULL ;
			pcm_scc_enable(&m_sccInfo, 0x00) ;
			return ;
		}

		// 繰り返し
		p_bgm = &m_p_bgmHead[(uint32_t )p_bgm[1]*5] ;
	}

	if( m_volume ) {
		if( !--m_volume ) {
			m_p_bgmHead = NULL ;

			pcm_scc_setReg(&m_sccInfo, 0, 0x0000) ;
			pcm_scc_setReg(&m_sccInfo, 1, 0x0000) ;
			pcm_scc_setReg(&m_sccInfo, 2, 0x0000) ;
			pcm_scc_setReg(&m_sccInfo, 3, 0x0000) ;
			pcm_scc_setReg(&m_sccInfo, 4, 0x0000) ;

			pcm_scc_enable(&m_sccInfo, 0x00) ;
			return ;
		}

		max = m_volume / 8 + 1 ;

		for(index=0; index<5; index++) {
			volume = p_bgm[index] >> 12 ;
			if( volume ) {
				volume = (p_bgm[index] & 0x0fff) | (((volume + 1) * max / 16) << 12) ;
			}
			data[index] = volume ;
		}

		pcm_scc_setReg(&m_sccInfo, 0, data[0]) ;
		pcm_scc_setReg(&m_sccInfo, 1, data[1]) ;
		pcm_scc_setReg(&m_sccInfo, 2, data[2]) ;
		pcm_scc_setReg(&m_sccInfo, 3, data[3]) ;
		pcm_scc_setReg(&m_sccInfo, 4, data[4]) ;

	} else {
		pcm_scc_setReg(&m_sccInfo, 0, p_bgm[0]) ;
		pcm_scc_setReg(&m_sccInfo, 1, p_bgm[1]) ;
		pcm_scc_setReg(&m_sccInfo, 2, p_bgm[2]) ;
		pcm_scc_setReg(&m_sccInfo, 3, p_bgm[3]) ;
		pcm_scc_setReg(&m_sccInfo, 4, p_bgm[4]) ;
	}

	m_p_bgm = &p_bgm[5] ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
void sound_init( void )
{
	int		index ;


	DAC_INIT() ;

	m_output = 0x0800 ;
	m_volume = 0 ;

	dpcm3_reset( &m_dpcm3Info ) ;
	pcm_scc_reset( &m_sccInfo ) ;

	for(index=0; index<=0x0800; index+=0x0040) {
		DAC_WRITE( index ) ;
		DAC_WAIT() ;
		sleep_us( 10 ) ;
	}

	add_repeating_timer_us(-SOUNDDAC_CFG_TIME, isr_timer, NULL, &m_timer) ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 処理                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_vsync( void )
{
	int		index ;

	// SCC 更新
	updateSCC() ;

	// DPCM 更新
	for(index=0; index<DPCM3_CFG_CHANNEL_COUNT; index++) {
		if( g.requestSE[index] ) {
			dpcm3_setData(&m_dpcm3Info, index, g_p_seData[g.requestSE[index]-1]) ;
			g.requestSE[index] = 0 ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  BGM 再生                                                                       
//!                                                                                 
//!  @param [in] no  BGM 番号                                                       
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_playBGM( int no )
{
	uint16_t	const *p_data ;


	m_p_bgmHead = NULL ;

	pcm_scc_enable(&m_sccInfo, 0x00) ;
	pcm_scc_reset( &m_sccInfo ) ;

	p_data = g_p_bgmData[no] ;

	pcm_scc_copyWaveData(&m_sccInfo, 0, (int8_t const *)&p_data[0], 128) ;

	m_p_bgmHead = &p_data[128/2] ;
	m_p_bgm     = &m_p_bgmHead[0] ;

	pcm_scc_enable(&m_sccInfo, 0x1f) ;
}


// -------------------------------------------------------------------------------- 
//!  BGM 停止                                                                       
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_stopBGM( void )
{
	m_p_bgmHead = NULL ;

	pcm_scc_enable(&m_sccInfo, 0x00) ;
}


// -------------------------------------------------------------------------------- 
//!  BGM フェードアウト                                                             
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_fadeOutBGM( void )
{
	if( m_p_bgmHead != NULL ) {
		m_volume = 15 * 8 ;
	}
}
