// -------------------------------------------------------------------------------- 
//!  @file  option.c                                                                
//!  オプション画面のソースファイル                                                 
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "option.h"


	enum {
		MENUITEM_TYPE_BUTTON,										//!< ボタン
		MENUITEM_TYPE_LRPARAM,										//!< 左右パラメータ
		MENUITEM_TYPE_PROC											//!< 処理
	} ;

	//! メニュー項目
	typedef struct {
		uint8_t						type ;							//!< メニュー項目種類
		uint8_t						x ;								//!< ｘ座標
		uint8_t						y ;								//!< ｙ座標
		uint8_t						count ;							//!< パラメータ数
		char						const *p_string ;				//!< 文字列
		void						*p_param ;						//!< パラメータ
	} MENUITEM ;

static void RAMFUNC option_040( SYSTEMTCB *p_tcb ) ;
static void RAMFUNC option_080( SYSTEMTCB *p_tcb ) ;
static void RAMFUNC option_0c0( SYSTEMTCB *p_tcb ) ;

//! メニュー項目
static const MENUITEM				m_menuItem[]={
										{ MENUITEM_TYPE_BUTTON,   8,  8, 0, "POWERUP",                           (void *)&g.cfg.button[0] },
										{ MENUITEM_TYPE_BUTTON,  11,  9, 0,    "SHOT",                           (void *)&g.cfg.button[1] },
										{ MENUITEM_TYPE_BUTTON,   8, 10, 0, "MISSILE",                           (void *)&g.cfg.button[2] },
										{ MENUITEM_TYPE_LRPARAM, 10, 12, 2,   "RAPID\0" " OFF>\0" "<ON  \0",     (void *)&g.cfg.rapid     },
										{ MENUITEM_TYPE_LRPARAM,  9, 15, 3,  "PLAYER\0" " 3>\0" "<5>\0" "<7 \0", (void *)&g.cfg.player    },
										{ MENUITEM_TYPE_PROC,    11, 24, 0,    "EXIT",                           (void *)option_080       }
									} ;


// -------------------------------------------------------------------------------- 
//!  パラメータ文字列                                                               
//!                                                                                 
//!  @param [in] p_string  文字列                                                   
//!  @param [in] param     パラメータ番号                                           
//!  @return               文字列                                                   
// -------------------------------------------------------------------------------- 
static char const *getParamString(char const *p_string, int param)
{
	do {
		while( *p_string++ != '\0' ) {
			;
		}
	} while( param-- ) ;

	return &p_string[0] ;
}


// -------------------------------------------------------------------------------- 
//!  オプション画面 VBlank 処理                                                     
//!                                                                                 
//!  @param [in] p_tcb  システムタスク TCB                                          
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static void RAMFUNC option_vblank( SYSTEMTCB *p_tcb )
{
	vga_setBgTile(&g.vram.tile[0][0], &g.vram.pat[0][0], &g.vram.pal[0][0]) ;
}


// -------------------------------------------------------------------------------- 
//!  オプション画面                                                                 
//!                                                                                 
//!  @param [in,out] p_tcb  システムタスク TCB                                      
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
void RAMFUNC option( SYSTEMTCB *p_tcb )
{
	MENUITEM	const *p_menuItem ;
	char		buff[6] ;
	int			index ;

	// 非表示
	vga_setBgBlank() ;
	vga_setFgBlank() ;
	vga_setTextOffset(0, 0) ;
	vga_setSpriteOffset(0, 0) ;
	vga_setTileOffset(0, 0) ;

	setVSyncProc(NULL, NULL) ;
	vga_setHSyncProc( NULL ) ;

	// 初期化
	setMemory(&g.vram.tile[0][0], 0, sizeof( g.vram.tile )) ;
	setMemory(&g.vram.pal[0][0],  0, sizeof( g.vram.pal )) ;
	p_tcb->cursor = 0 ;

	// データ転送
	xferData(XFERDATA_FIX_OFFSET, XFERDATA_FIX_COUNT) ;

	// 文字列描画
	writeTileString16(14, 1, "OPTION", 0) ;

	// メニュー項目描画
	p_menuItem = &m_menuItem[0] ;
	for(index=0; index<sizeof( m_menuItem )/sizeof( m_menuItem[0] ); index++,p_menuItem++) {
		writeTileString8(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], index==0) ;

		switch( p_menuItem->type ) {
			case MENUITEM_TYPE_BUTTON:
			{
				sprintf(&buff[0], "B%d ", *(uint8_t *)p_menuItem->p_param) ;
				writeTileString8(17, p_menuItem->y, &buff[0], 0) ;
				break ;
			}

			case MENUITEM_TYPE_LRPARAM:
			{
				writeTileString8(16, p_menuItem->y, getParamString(&p_menuItem->p_string[0], *(uint8_t *)p_menuItem->p_param), 0) ;
				break ;
			}

			default:
				break ;
		}
	}

	setVSyncProc(option_vblank, p_tcb) ;

	// タスク破棄
	task_destroyPrio(TASK_PRIO_SYSTEM+1, TASK_PRIO_POST-1) ;

	p_tcb->taskProc = option_040 ;
}


static void RAMFUNC option_040( SYSTEMTCB *p_tcb )
{
	MENUITEM	const *p_menuItem ;
	uint8_t		cursor ;


	p_menuItem = &m_menuItem[p_tcb->cursor] ;

	// カーソル移動
	if( g.pad.once & (PAD_SELECT|PAD_UP|PAD_DOWN) ) {
		cursor = p_tcb->cursor ;
		if( g.pad.once & PAD_UP ) {
			if( --p_tcb->cursor == (uint8_t )-1 ) {
				p_tcb->cursor = sizeof( m_menuItem )/sizeof( m_menuItem[0] ) - 1 ;
			}
		} else {
			if( ++p_tcb->cursor == (sizeof( m_menuItem )/sizeof( m_menuItem[0] )) ) {
				p_tcb->cursor = 0 ;
			}
		}

		if( p_tcb->cursor != cursor ) {
			writeTileString8(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], 0) ;

			p_menuItem = &m_menuItem[p_tcb->cursor] ;
			writeTileString8(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], 1) ;

			sound_requestSE(0, SE_00) ;
		}
		return ;
	}

	switch( p_menuItem->type ) {
		case MENUITEM_TYPE_BUTTON:
		{
			char		buff[8] ;
			uint8_t		*p_button ;
			int			index ;

			// ボタン番号取得
			for(index=PAD_BTN_; index<=PAD_RB_; index++) {
				if( g.pad.norm & PAD_BTN( index ) ) {
					break ;
				}
			}
			if( index > PAD_RB_ ) {
				break ;
			}

			p_button = (uint8_t *)p_menuItem->p_param ;
			if( *p_button == index ) {
				break ;
			}

			*p_button = (uint8_t )index ;

			sprintf(&buff[0], "B%d ", index) ;
			writeTileString8(17, p_menuItem->y, &buff[0], 0) ;
			break ;
		}

		case MENUITEM_TYPE_LRPARAM:
		{
			if( !(g.pad.once&(PAD_LEFT|PAD_RIGHT)) ) {
				break ;
			}

			cursor = *(uint8_t *)p_menuItem->p_param ;

			if( g.pad.once & PAD_LEFT ) {
				if( cursor ) {
					cursor-- ;
				}
			} else {
				if( cursor < (p_menuItem->count-1) ) {
					cursor++ ;
				}
			}

			if( *(uint8_t *)p_menuItem->p_param == cursor ) {
				break ;
			}
			*(uint8_t *)p_menuItem->p_param = cursor ;

			// パラメータ描画
			writeTileString8(16, p_menuItem->y, getParamString(&p_menuItem->p_string[0], cursor), 0) ;

			sound_requestSE(0, SE_00) ;
			break ;
		}

		case MENUITEM_TYPE_PROC:
		{
			if( g.pad.once & (PAD_START|PAD_BTN( 0 )|PAD_BTN( 1 )|PAD_BTN( 2 )) ) {
				((TASKPROC )p_menuItem->p_param)( p_tcb ) ;
			}
			break ;
		}
	}
}


static void RAMFUNC option_080( SYSTEMTCB *p_tcb )
{
	sound_requestSE(0, SE_02) ;

	p_tcb->taskProc = option_0c0 ;
	p_tcb->timer0   = 90 ;
}


static void RAMFUNC option_0c0( SYSTEMTCB *p_tcb )
{
	MENUITEM	const *p_menuItem ;


	p_menuItem = &m_menuItem[p_tcb->cursor] ;

	if( !--p_tcb->timer0 ) {
		task_setSystemMode( SYSTEM_MODE_TITLE ) ;
		return ;
	}

	if( g.count1 & 7 ) {
		return ;
	}

	if( p_tcb->timer0 & 8 ) {
		writeTileString8(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], 1) ;
	} else {
		writeTileString8(p_menuItem->x, p_menuItem->y, "          ",             1) ;
	}
}
