// -------------------------------------------------------------------------------- 
//!  @file  task.c                                                                  
//!  タスク管理のソースファイル                                                     
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "system/title.h"
#include "system/powermeterselect.h"
#include "system/option.h"
#include "system/game.h"


static SYSTEMTCB					m_systemTcb ;					//!< システムタスク TCB
static POSTTCB						m_postTcb ;						//!< 後処理タスク TCB
static TCB							m_tcb[TASK_MAX] ;				//!< タスクコントロールブロック
static TCB							*m_p_nonExistentTcb ;			//!< 未登録状態タスクリスト
static TCB							*m_p_dormantTcb ;				//!< 休止状態タスクリスト
static TCB							*m_p_nextTcb ;					//!< 次の実行タスク


// -------------------------------------------------------------------------------- 
//!  タスクを取り付け                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  取り付けタスク TCB                                      
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC attach( TCB *p_tcb )
{
	TCB			*p_prev, *p_next ;
	uint8_t		prio ;


	prio = p_tcb->prio ;

	if( prio < TASK_PRIO_ENEMY0 ) {		// 先頭から取り付け

		p_next = m_systemTcb.p_next ;
		while( prio > p_next->prio ) {
			p_next = p_next->p_next ;
		}

		p_tcb->p_next          = p_next ;
		p_tcb->p_prev          = p_next->p_prev ;
		p_next->p_prev->p_next = p_tcb ;
		p_next->p_prev         = p_tcb ;

	} else {							// 最後から取り付け

		p_prev = m_postTcb.p_prev ;
		while( prio < p_prev->prio ) {
			p_prev = p_prev->p_prev ;
		}

		p_tcb->p_prev          = p_prev ;
		p_tcb->p_next          = p_prev->p_next ;
		p_prev->p_next->p_prev = p_tcb ;
		p_prev->p_next         = p_tcb ;

	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void task_init( void )
{
	int		index ;

	// 初期化
	m_p_dormantTcb = NULL ;

	// 未登録状態タスクリスト作成
	m_p_nonExistentTcb = &m_tcb[0] ;
	for(index=0; index<TASK_MAX; index++) {
		m_tcb[index].p_next = &m_tcb[index+1] ;
		m_tcb[index].status = TCB_STATUS_NONEXISTENT ;
	}

	// システムタスク
	m_systemTcb.p_next = (TCB *)&m_postTcb ;
	m_systemTcb.prio   = TASK_PRIO_SYSTEM ;

	// 後処理タスク
	m_postTcb.p_prev   = (TCB *)&m_systemTcb ;
	m_postTcb.prio     = TASK_PRIO_POST ;

	task_setSystemMode( SYSTEM_MODE_TITLE ) ;
}


// -------------------------------------------------------------------------------- 
//!  タスクを生成                                                                   
//!                                                                                 
//!  @param [in] p_parent  親タスク TCB                                             
//!  @param [in] taskProc  タスク処理                                               
//!  @param [in] prio      優先順位                                                 
//!  @retval     NULL      TCB 不足                                                 
//!  @retval     上記以外  TCB アドレス                                             
// -------------------------------------------------------------------------------- 
void * RAMFUNC task_create(TCB *p_parent, TASKPROC taskProc, uint8_t prio)
{
	TCB		*p_tcb ;

	// TCB 取得
	p_tcb = m_p_nonExistentTcb ;
	if( p_tcb == NULL ) {
		DBGTRACE("**** %s(%d)\n", __func__, __LINE__) ;
		return NULL ;
	}
	m_p_nonExistentTcb = p_tcb->p_next ;

	// 初期化
	p_tcb->p_parent  = p_parent ;
	p_tcb->taskProc  = taskProc ;
	p_tcb->pauseProc = task_sleep ;
	p_tcb->status    = TCB_STATUS_DORMANT ;
	p_tcb->prio      = prio ;

	// 休止状態タスクリストへ追加
	p_tcb->p_next  = m_p_dormantTcb ;
	m_p_dormantTcb = p_tcb ;

	return p_tcb ;
}


// -------------------------------------------------------------------------------- 
//!  タスクを破棄                                                                   
//!                                                                                 
//!  @param [in,out] p_tcb  TCB アドレス                                            
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
void RAMFUNC task_destroy( TCB *p_tcb )
{
	TCB		*p_prev, *p_next ;


	if( p_tcb == m_p_nextTcb ) {
		m_p_nextTcb = p_tcb->p_next ;
	}

	// TCB 取り外し
	p_prev         = p_tcb->p_prev ;
	p_next         = p_tcb->p_next ;
	p_prev->p_next = p_next ;
	p_next->p_prev = p_prev ;

	// 未登録状態タスクリストへ追加
	p_tcb->status      = TCB_STATUS_NONEXISTENT ;
	p_tcb->p_next      = m_p_nonExistentTcb ;
	m_p_nonExistentTcb = p_tcb ;
}


// -------------------------------------------------------------------------------- 
//!  優先順位によるタスク破棄                                                       
//!                                                                                 
//!  @param [in] first  最初の優先順位                                              
//!  @param [in] last   最後の優先順位                                              
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
void RAMFUNC task_destroyPrio(uint8_t first, uint8_t last)
{
	TCB		*p_tcb, *p_next, **pp_next ;

	// 先頭 TCB 取得
	p_tcb = m_systemTcb.p_next ;
	while( p_tcb->prio < first ) {
		p_tcb = p_tcb->p_next ;
		if( p_tcb == NULL ) {
			return ;
		}
	}

	// 破棄
	while( p_tcb->prio <= last ) {
		p_next = p_tcb->p_next ;

		task_destroy( p_tcb ) ;

		p_tcb = p_next ;
		if( p_tcb == NULL ) {
			return ;
		}
	}

	// 休止状態タスク破棄
	pp_next = &m_p_dormantTcb ;
	while( 1 ) {
		p_tcb = *pp_next ;
		if( p_tcb == NULL ) {
			break ;
		}

		if( (p_tcb->prio>=first) || (p_tcb->prio<=last) ) {

			// 破棄
			*pp_next = p_tcb->p_next ;

			// 未登録状態タスクリスト
			p_tcb->status      = TCB_STATUS_NONEXISTENT ;
			p_tcb->p_next      = m_p_nonExistentTcb ;
			m_p_nonExistentTcb = p_tcb ;

		} else {
			pp_next = &p_tcb->p_next ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  指定優先度の TCB を取得                                                        
//!                                                                                 
//!  @param [in] prio      優先順位                                                 
//!  @retval     NULL      タスクなし                                               
//!  @retval     上記以外  TCB アドレス                                             
// -------------------------------------------------------------------------------- 
void * RAMFUNC task_getTcbPrio( uint8_t prio )
{
	TCB		*p_tcb ;


	p_tcb = m_systemTcb.p_next ;

	while( p_tcb != NULL ) {
		if( p_tcb->prio >= prio ) {
			if( p_tcb->prio != prio ) {
				break ;
			}
			return p_tcb ;
		}

		p_tcb = p_tcb->p_next ;
	}

	return NULL ;
}


// -------------------------------------------------------------------------------- 
//!  スリープ                                                                       
//!                                                                                 
//!  @param [in] p_tcb  TCB アドレス                                                
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
void RAMFUNC task_sleep( TCB *p_tcb )
{
}


// -------------------------------------------------------------------------------- 
//!  タスク実行                                                                     
//!                                                                                 
//!  @param [in] pause  - true:停止処理                                             
//!                     - false:通常処理                                            
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC task_exec( bool pause )
{
	TCB		*p_tcb ;


	hit_init() ;

	g.task = 1 ;

	// システムタスク実行
	m_systemTcb.taskProc( (TCB *)&m_systemTcb ) ;
	p_tcb = m_systemTcb.p_next ;

	if( pause == false ) {
		do {
			m_p_nextTcb = p_tcb->p_next ;
			p_tcb->taskProc( p_tcb ) ;
			p_tcb = m_p_nextTcb ;
		} while( p_tcb != NULL ) ;
	} else {
		do {
			m_p_nextTcb = p_tcb->p_next ;
			p_tcb->pauseProc( p_tcb ) ;
			p_tcb = m_p_nextTcb ;
		} while( p_tcb != NULL ) ;
	}

	g.task = 0 ;

	hit_calc() ;

	// 休止状態タスクリストからレディーリストへ取り付け
	while( 1 ) {
		p_tcb = m_p_dormantTcb ;
		if( p_tcb == NULL ) {
			break ;
		}

		m_p_dormantTcb = p_tcb->p_next ;
		p_tcb->status &= ~TCB_STATUS_DORMANT ;

		attach( p_tcb ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  タスク状態をコンソールへ出力                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
int RAMFUNC task_debug( void )
{
	TCB		*p_tcb ;
	int		index ;


	printf("non existent:\n") ;

	p_tcb = m_p_nonExistentTcb ;
	index = 0 ;
	while( p_tcb ) {
		printf("  %3d: %p\n", index++, p_tcb) ;

		p_tcb = p_tcb->p_next ;
	}

	printf("\nready:\n") ;

	p_tcb = (TCB *)&m_systemTcb ;
	index = 0 ;
	do {
		printf("  %3d: %p(%d)\n", index++, p_tcb, p_tcb->prio) ;

		p_tcb = p_tcb->p_next ;
	} while( p_tcb != NULL ) ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  システムモードを設定                                                           
//!                                                                                 
//!  @param [in] mode  システムモード                                               
// -------------------------------------------------------------------------------- 
void task_setSystemMode( int mode )
{
	static const TASKPROC	taskProc[][2]={
								{ title,            task_sleep },	// SYSTEM_MODE_TITLE
								{ powermeterselect, task_sleep },	// SYSTEM_MODE_POWERMETERSELECT
								{ option,           task_sleep },	// SYSTEM_MODE_OPTION
								{ game,             task_sleep },	// SYSTEM_MODE_GAME
								{ game_miss,        task_sleep },	// SYSTEM_MODE_MISS
								{ game_gameover,    task_sleep },	// SYSTEM_MODE_GAMEOVER
							} ;


	m_systemTcb.taskProc = taskProc[mode][0] ;
	m_postTcb.taskProc   = taskProc[mode][1] ;
}
