// -------------------------------------------------------------------------------- 
//!  @file  wavedpcm3.c                                                             
//!  wave ファイルから 3bit DPCM データの変換ツール                                 
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


	enum {
		DELTA_BIT					=	3,							//!< デルタデータビット数
		DELTA_COUNT					=	1 << DELTA_BIT,				//!< デルタデータ数
		CHUNK_ID_RIFF				=	0x46464952,					//!< RIFF チャンク ID
		CHUNK_ID_FMT				=	0x20746d66,					//!< fmt  チャンク ID
		CHUNK_ID_DATA				=	0x61746164,					//!< data チャンク ID
		RIFF_FORMAT					=	0x45564157,					//!< RIFF チャンクフォーマット
		WAVEDATA_MAX				=	65535,						//!< 波形データ最大サイズ
	} ;

	//! ファイル情報
	typedef struct {
		FILE						*p_file ;						//!< ファイルポインタ
		uint32_t					size ;							//!< サイズ
		uint32_t					pos ;							//!< 位置
	} FILEINFO ;

	//! チャンク
	typedef struct {
		uint32_t					id ;							//!< チャンク ID
		uint32_t					size ;							//!< チャンクサイズ
	} CHUNK ;

	//! 波形フォーマット
	typedef struct {
		uint16_t					formatTag ;						//!< フォーマットタグ
		uint16_t					channels ;						//!< チャンネル数
		uint32_t					samplesPerSec ;					//!< サンプリング周波数
		uint32_t					avgBytesPerSec ;				//!< １秒あたりのバイト数
		uint16_t					blockAlign ;					//!< ブロックサイズ
		uint16_t					bitsPerSample ;					//!< サンプルビット数
	} WAVEFORMAT ;


// -------------------------------------------------------------------------------- 
//!  サンプル数から16bitにアラインメントされたバイト数へ変換                        
//!                                                                                 
//!  @param [in] sample  サンプル数                                                 
//!  @return             バイト数                                                   
// -------------------------------------------------------------------------------- 
static inline int sampleToSize( uint32_t sample )
{
	return ((sample * DELTA_BIT + 15) / 16 * 2) ;
}


// -------------------------------------------------------------------------------- 
//!  ファイル読み込み                                                               
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @param [out]    p_buff      読み込みバッファアドレス                           
//!  @param [in]     size        読み込みバッファサイズ                             
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int readFile(FILEINFO *p_fileInfo, void *p_buff, uint32_t size)
{
	if( (p_fileInfo->pos+size) > p_fileInfo->size ) {
		return -1 ;
	}

	if( fread(p_buff, size, 1, p_fileInfo->p_file) != 1 ) {
		return -1 ;
	}

	p_fileInfo->pos += size ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ファイルシーク                                                                 
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @param [in]     pos         シーク位置                                         
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int seekFile(FILEINFO *p_fileInfo, uint32_t pos)
{
	if( p_fileInfo->pos == pos ) {
		return 0 ;
	}

	if( p_fileInfo->size < pos ) {
		return -1 ;
	}

	if( fseek(p_fileInfo->p_file, pos, SEEK_SET) == -1 ) {
		return -1 ;
	}

	p_fileInfo->pos = pos ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  文字列からデルタデータを取得                                                   
//!                                                                                 
//!  @param [in]  p_string  文字列                                                  
//!  @param [out] p_delta   デルタデータ格納アドレス                                
//!  @retval       0        正常終了                                                
//!  @retval      -1        異常終了                                                
// -------------------------------------------------------------------------------- 
static int getDeltaData(const char *p_string, int8_t *p_deltaData)
{
	char	sign, eoc ;
	int		index, value ;


	eoc = ',' ;

	for(index=0; index<DELTA_COUNT; index++) {
		if( index == 7 ) {
			eoc = 0 ;
		}

		// 符号
		sign = 0 ;
		if( *p_string == '-' ) {
			sign = 1 ;
			p_string++ ;
		}

		// 値
		value = 0 ;

		while( 1 ) {
			if( !isdigit( *p_string ) ) {
				return -1 ;
			}

			value = (value * 10) + (*p_string++ - '0') ;
			if( value > 128 ) {
				return -1 ;
			}

			if( *p_string == eoc ) {
				p_string += (eoc == ',') ;
				break ;
			}

			if( *p_string == '\0' ) {
				return -1 ;
			}
		}

		if( sign ) {
			value = -value ;
		}

		if( value > 127 ) {
			return -1 ;
		}

		*p_deltaData++ = value ;
	}

	if( *p_string != '\0' ) {
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  RIFF チャンクを確認                                                            
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int checkRiffChunk( FILEINFO *p_fileInfo )
{
	CHUNK		chunk ;
	uint32_t	format ;


	if( p_fileInfo->size < sizeof( CHUNK ) ) {
		printf("サポートしていないファイル形式です。\n") ;
		return -1 ;
	}

	// RIFF チャンク読み込み
	if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
		printf("RIFFチャンクの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( chunk.id != CHUNK_ID_RIFF ) {
		printf("RIFFチャンクがありません。\n") ;
		return -1 ;
	}

	if( chunk.size < (p_fileInfo->size-sizeof( CHUNK )) ) {
		printf("RIFFチャンクのサイズが正しくありません。\n") ;
		return -1 ;
	}

	p_fileInfo->size = sizeof( CHUNK ) + chunk.size ;

	if( readFile(p_fileInfo, &format, sizeof( format )) ) {
		printf("RIFFチャンクのフォーマットの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( format != 0x45564157 ) {		// 0x45564157 = "WAVE"
		printf("サポートしていないファイル形式です。\n") ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  波形フォーマットを読み込む                                                     
//!                                                                                 
//!  @param [in,out] p_fileInfo    ファイル情報                                     
//!  @param [out]    p_waveFormat  波形フォーマット格納アドレス                     
//!  @retval          0            正常終了                                         
//!  @retval         -1            異常終了                                         
// -------------------------------------------------------------------------------- 
static int readWaveFormat(FILEINFO *p_fileInfo, WAVEFORMAT *p_waveFormat)
{
	CHUNK	chunk ;
	int		remain ;

	// fmt チャンク読み込み
	if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
		printf("fmt チャンクの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( chunk.id != 0x20746d66 ) {		// 0x20746d66 = "fmt "
		printf("fmt チャンクがありません。\n") ;
		return -1 ;
	}

	if( chunk.size < sizeof( WAVEFORMAT ) ) {
		printf("fmt チャンクのサイズが正しくありません。\n") ;
		return -1 ;
	}

	if( readFile(p_fileInfo, p_waveFormat, sizeof( WAVEFORMAT )) ) {
		printf("フォーマットの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	remain = ((chunk.size + 1) & ~1) - sizeof( WAVEFORMAT ) ;
	if( remain ) {
		if( seekFile(p_fileInfo, p_fileInfo->pos+remain) ) {
			printf("フォーマットの読み込みに失敗しました。\n") ;
			return -1 ;
		}
	}

	printf("fmt:\n") ;
	printf("    formatTag      = 0x%04x\n", p_waveFormat->formatTag) ;
	printf("    channels       = %d\n",     p_waveFormat->channels) ;
	printf("    samplesPerSec  = %d\n",     p_waveFormat->samplesPerSec) ;
	printf("    avgBytesPerSec = %d\n",     p_waveFormat->avgBytesPerSec) ;
	printf("    blockAlign     = %d\n",     p_waveFormat->blockAlign) ;
	printf("    bitsPerSample  = %d\n",     p_waveFormat->bitsPerSample) ;

	if( p_waveFormat->formatTag != 0x0001 ) {
		printf("PCM 形式以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( p_waveFormat->channels != 1 ) {
		printf("モノラル以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( p_waveFormat->bitsPerSample != 8 ) {
		printf("8ビット以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( !p_waveFormat->samplesPerSec || (p_waveFormat->avgBytesPerSec!=p_waveFormat->samplesPerSec) || (p_waveFormat->blockAlign!=1) ) {
		printf("フォーマットが正しくありません。\n") ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  波形データを読み込む                                                           
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @param [out]    p_waveData  波形データ格納アドレス                             
//!  @retval          0 以上     波形データサイズ                                   
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int readWaveData(FILEINFO *p_fileInfo, uint8_t *p_waveData)
{
	CHUNK	chunk ;


	while( 1 ) {
		// チャンク読み込み
		if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
			printf("dataチャンクの読み込みに失敗しました。\n") ;
			return -1 ;
		}

		if( chunk.id == CHUNK_ID_DATA ) {
			break ;
		}

		if( seekFile(p_fileInfo, p_fileInfo->pos+((chunk.size+1)&~1)) ) {
			printf("dataチャンクの読み込みに失敗しました。\n") ;
			return -1 ;
		}
	}

	if( !chunk.size || ((p_fileInfo->pos+chunk.size)>p_fileInfo->size) ) {
		printf("dataチャンクのサイズ（%d）が正しくありません。\n", chunk.size) ;
		return -1 ;
	}

	if( chunk.size >= WAVEDATA_MAX ) {
		printf("波形データが%dを超えています（%d）。\n", WAVEDATA_MAX, chunk.size) ;
	}

	if( readFile(p_fileInfo, &p_waveData[0], chunk.size) ) {
		printf("dataチャンクの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	printf("\ndata:\n") ;
	printf("    sample = %d\n", chunk.size) ;

	return chunk.size ;
}


// -------------------------------------------------------------------------------- 
//!  波形データをロード                                                             
//!                                                                                 
//!  @param [in]  p_fileName  ファイル名                                            
//!  @param [out] p_wave      波形データ格納アドレス                                
//!  @retval       0 以上     サンプル数                                            
//!  @retval      -1          異常終了                                              
// -------------------------------------------------------------------------------- 
static int loadWaveData(char const *p_fileName, uint8_t *p_waveData)
{
	FILEINFO	fileInfo ;
	WAVEFORMAT	waveFormat ;
	int			result ;
	long		fileSize ;

	// ファイル開く
	fileInfo.p_file = fopen(&p_fileName[0], "rb") ;
	if( fileInfo.p_file == NULL ) {
		fprintf(stderr, "%sファイルを開くのに失敗しました。\n", &p_fileName[0]) ;
		return -1 ;
	}

	result = -1 ;

	do {
		// ファイルサイズ取得
		fseek(fileInfo.p_file, 0, SEEK_END) ;

		fileSize = ftell( fileInfo.p_file ) ;
		if( fileSize <= 0 ) {
			break ;
		}
		fileInfo.size = (uint32_t )fileSize ;

		fseek(fileInfo.p_file, 0, SEEK_SET) ;

		fileInfo.pos = 0 ;

		// RIFF チャンク確認
		if( checkRiffChunk( &fileInfo ) ) {
			break ;
		}

		// 波形フォーマット読み込み
		if( readWaveFormat(&fileInfo, &waveFormat) ) {
			break ;
		}

		// 波形データ読み込み
		result = readWaveData(&fileInfo, &p_waveData[0]) ;
		if( result < 0 ) {
			break ;
		}
	} while( 0 ) ;

	fclose( fileInfo.p_file ) ;

	return result ;
}


// -------------------------------------------------------------------------------- 
//!  波形データからデルタテーブルインデックスデータを作成                           
//!                                                                                 
//!  @param [in] p_waveData    波形データ                                           
//!  @param [in] sample        サンプル数                                           
//!  @param [in] p_deltaData   デルタデータ                                         
//!  @param [in] p_deltaIndex  デルタデータインデックスデータ格納アドレス           
//!  @return                   なし                                                 
// -------------------------------------------------------------------------------- 
static void createDeltaIndex(uint8_t *p_waveData, int sample, int8_t *p_deltaData, uint8_t *p_deltaIndex)
{
	uint8_t		srce, dest, deltaIndex, *p_waveDataEnd ;
	uint32_t	count[DELTA_COUNT] ;
	int			index, diff, calc, diffCount ;


	memset(&count[0], 0, sizeof( count )) ;
	diffCount = 0 ;

	// デルタテーブルインデックスデータ
	p_waveDataEnd = &p_waveData[sample] ;
	dest          = 0x80 ;

	do {
		srce       = *p_waveData++ ;
		diff       = 256 ;
		deltaIndex = 0 ;

		for(index=0; index<8; index++) {
			calc = (dest + p_deltaData[index]) - srce ;
			if( !calc ) {
				diff       = 0 ;
				deltaIndex = index ;
				break ;
			}

			if( calc < 0 ) {
				calc = -calc ;
			}

			if( diff > calc ) {
				diff       = calc ;
				deltaIndex = index ;
			}
		}

		diffCount += diff ;

		dest           += p_deltaData[deltaIndex] ;
		*p_deltaIndex++ = deltaIndex ;

		count[deltaIndex]++ ;
	} while( p_waveData != p_waveDataEnd ) ;

	// 出現数表示
	printf("\ndelta count: %d\n", diffCount) ;
	for(index=0; index<DELTA_COUNT; index++) {
		printf("    %d = %d\n", index, count[index]) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  デルタテーブルインデックスデータ から DPCM データを作成                        
//!                                                                                 
//!  @param [in]  p_deltaIndex  デルタテーブルインデックスデータ                    
//!  @param [in]  sample        サンプル数                                          
//!  @param [out] p_dpcmData    DPCM データ格納アドレス                             
//!  @return                    なし                                                
// -------------------------------------------------------------------------------- 
static void createDpcmData(uint8_t *p_deltaIndex, int sample, uint16_t *p_dpcmData)
{
	uint8_t		*p_dest ;
	uint32_t	dpcmData ;
	int			index ;


	p_dest = (uint8_t *)&p_dpcmData[0] ;
	for(index=0; index<sample; index+=8) {
		dpcmData      = ((uint32_t )p_deltaIndex[0]<<(3*0)) |
		                ((uint32_t )p_deltaIndex[1]<<(3*1)) |
		                ((uint32_t )p_deltaIndex[2]<<(3*2)) |
		                ((uint32_t )p_deltaIndex[3]<<(3*3)) |
		                ((uint32_t )p_deltaIndex[4]<<(3*4)) |
		                ((uint32_t )p_deltaIndex[5]<<(3*5)) |
		                ((uint32_t )p_deltaIndex[6]<<(3*6)) |
		                ((uint32_t )p_deltaIndex[7]<<(3*7)) ;
		p_deltaIndex += 8 ;

		p_dest[0] = (uint8_t )(dpcmData>> 0) ;
		p_dest[1] = (uint8_t )(dpcmData>> 8) ;
		p_dest[2] = (uint8_t )(dpcmData>>16) ;
		p_dest   += DELTA_BIT ;
	}
}


// -------------------------------------------------------------------------------- 
//!  DPCM ファイルを作成                                                            
//!                                                                                 
//!  @param [in] p_fileName  ファイル名                                             
//!  @param [in] p_dpcmData  DPCM データ                                            
//!  @param [in] sample      サンプル数                                             
//!  @param [in] p_dpcmData  デルタデータ                                           
//!  @retval      0          正常終了                                               
//!  @retval     -0          異常終了                                               
// -------------------------------------------------------------------------------- 
static int createDpcmFile(char const *p_fileName, uint16_t *p_dpcmData, int sample, int8_t *p_deltaData)
{
	FILE		*p_file ;
	uint16_t	data ;
	int			result ;


	p_file = fopen(&p_fileName[0], "wb") ;
	if( p_file == NULL ) {
		return -1 ;
	}

	result = -1 ;

	do {
		// ヘッダー書き込み
		if( fwrite("DPCM", 4, 1, p_file) != 1 ) {
			break ;
		}

		// サンプル数書き込み
		data = (uint16_t )sample ;
		if( fwrite(&data, sizeof( data ), 1, p_file) != 1 ) {
			break ;
		}

		// デルタデータ書き込み
		if( fwrite(&p_deltaData[0], DELTA_COUNT, 1, p_file) != 1 ) {
			break ;
		}

		// DPCM データ書き込み
		if( fwrite(&p_dpcmData[0], sampleToSize( sample ), 1, p_file) != 1 ) {
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	result |= fclose( p_file ) ;

	if( result ) {
		remove( &p_fileName[0] ) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  DPCM データから波形データを作成                                                
//!                                                                                 
//!  @param [in]  p_dpcmData   DPCM データ                                          
//!  @param [in]  sample       サンプル数                                           
//!  @param [in]  p_deltaData  デルタデータ                                         
//!  @param [out] p_waveData   波形データ格納アドレス                               
//!  @retval       0           正常終了                                             
//!  @retval      -0           異常終了                                             
// -------------------------------------------------------------------------------- 
static void createWaveData(uint16_t *p_dpcmData, uint16_t sample, int8_t *p_deltaData, uint8_t *p_waveData)
{
	uint8_t		waveData ;
	uint32_t	dpcmData ;
	int			index, count ;


	dpcmData = *p_dpcmData++ ;
	count    = 16 ;
	waveData = 0x80 ;

	for(index=0; index<sample; index++) {
		if( count < DELTA_BIT ) {
			dpcmData |= ((uint32_t )*p_dpcmData++ << count) ;
			count    += 16 ;
		}

		waveData  += p_deltaData[dpcmData&7] ;
		dpcmData >>= DELTA_BIT ;
		count     -= DELTA_BIT ;

		*p_waveData++ = waveData ;
	}
}


// -------------------------------------------------------------------------------- 
//!  DPCM データから波形ファイルを作成                                              
//!                                                                                 
//!  @param [in] p_fileName   ファイル名                                            
//!  @param [in] p_dpcmData   DPCM データ                                           
//!  @param [in] sample       サンプル数                                            
//!  @param [in] p_deltaData  デルタデータ                                          
//!  @retval      0           正常終了                                              
//!  @retval     -0           異常終了                                              
// -------------------------------------------------------------------------------- 
static int createWaveFile(char const *p_fileName, uint16_t *p_dpcmData, int sample, int8_t *p_deltaData)
{
	FILE		*p_file ;
	CHUNK		chunk ;
	WAVEFORMAT	waveFormat ;
	uint8_t		waveData[65536] ;
	uint32_t	data ;
	int			result ;


	p_file = fopen(&p_fileName[0], "wb") ;
	if( p_file == NULL ) {
		return -1 ;
	}

	result = -1 ;

	do {
		// RIFF チャンク書き込み
		chunk.id   = CHUNK_ID_RIFF ;
		chunk.size = sizeof( data ) +
		             sizeof( chunk ) + sizeof( waveFormat ) +
		             sizeof( chunk ) + sample ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}

		data = RIFF_FORMAT ;
		if( fwrite(&data, sizeof( data ), 1, p_file) != 1 ) {
			break ;
		}

		// fmt チャンク書き込み
		chunk.id   = CHUNK_ID_FMT ;
		chunk.size = sizeof( WAVEFORMAT ) ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}

		waveFormat.formatTag      = 0x0001 ;
		waveFormat.channels       = 1 ;
		waveFormat.samplesPerSec  = 20000 ;
		waveFormat.avgBytesPerSec = 20000 ;
		waveFormat.blockAlign     = 1 ;
		waveFormat.bitsPerSample  = 8 ;
		if( fwrite(&waveFormat, sizeof( waveFormat ), 1, p_file) != 1 ) {
			break ;
		}

		// data チャンク書き込み
		chunk.id   = CHUNK_ID_DATA ;
		chunk.size = sample ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}

		createWaveData(&p_dpcmData[0], sample, &p_deltaData[0], &waveData[0]) ;

		if( fwrite(&waveData[0], sample, 1, p_file) != 1 ) {
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	result |= fclose( p_file ) ;

	if( result ) {
		remove( &p_fileName[0] ) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  メイン処理                                                                     
//!                                                                                 
//!  @param [in] argc  引数の数                                                     
//!  @param [in] argv  引数                                                         
//!  @retval     0     正常終了                                                     
//!  @retval     1     異常終了                                                     
//!  @retval     2     ヘルプ表示                                                   
// -------------------------------------------------------------------------------- 
int main(int argc, char *argv[])
{
	int8_t		deltaData[DELTA_COUNT] ;
	uint8_t		waveData[WAVEDATA_MAX], deltaIndex[WAVEDATA_MAX] ;
	uint16_t	dpcmData[WAVEDATA_MAX/2] ;
	int			sample ;


	if( argc < 4 ) {
		fprintf(stderr, "usage: wavedpcm3 wavefile delta dpcmfile [wavefile]\n"
		                "\n"
		                " delta  カンマ区切りの8個の数値（-128 - 127）\n") ;
		return 2 ;
	}

	// デルタデータ取得
	if( getDeltaData(argv[2], &deltaData[0]) ) {
		printf("デルタデータが正しくありません。\n") ;
		return 1 ;
	}

	memset(&waveData[0], 0x80, sizeof( waveData )) ;
	memset(&dpcmData[0], 0,    sizeof( dpcmData )) ;

	// 波形データロード
	sample = loadWaveData(argv[1], &waveData[0]) ;
	if( sample < 0 ) {
		return 1 ;
	}

	// デルタテーブルインデックスデータ作成
	createDeltaIndex(&waveData[0], sample, &deltaData[0], &deltaIndex[0]) ;

	// DPCM データ作成
	createDpcmData(&deltaIndex[0], sample, &dpcmData[0]) ;

	// DPCM ファイル作成
	if( createDpcmFile(argv[3], &dpcmData[0], sample, &deltaData[0]) ) {
		fprintf(stderr, "%s ファイルの生成に失敗しました。\n", argv[2]) ;
		return 1 ;
	}

	// 波形ファイル作成
	if( argc == 5 ) {
		if( createWaveFile(argv[4], &dpcmData[0], sample, &deltaData[0]) ) {
			fprintf(stderr, "%s ファイルの生成に失敗しました。\n", argv[2]) ;
			return 1 ;
		}
	}

	return 0 ;
}
